# Copyright ESIEE (2013) 
# 
# benjamin.perret@esiee.fr
# 
# This software is is an implementation of the image analysis method presented 
# in the research paper "Directed-connected hierarchical operators" 
# written by Benjamin Perret, Jean Cousty, Olena Tankyevych, Hugues Talbot, and Nicolas Passat
# 
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.




#####################################################################################################
#
# This software is a supplementary material for the article:
#         "Directed connected operators: Asymmetric hierarchies for image filtering and segmentation" 
# written by Benjamin Perret, Jean Cousty, Olena Tankyevych, Hugues Talbot, and Nicolas Passat
#
# This software is written for reproductible research. It contains an implementation of the algorithms 
# presented in the paper and the functions used to produce the results presented in the Section "Illustration". 
#
# This software is a package of 30 files:
#    - DirectedComponentHierarchy.py: the source code file writtten in python 3
#    - README.txt: a plain text file with some usefull information to start with 
#    - fig5-Graph.list: a plain text file encoding the graph of Fig. 5
#    - fig5-Image.png: a png image encoding the weights of the graph of Fig. 5
#    - heart.pgm: a pgm image of a heart RMI : Fig. 11 a
#    - heart-BackgroundMarker.pgm: a pgm image of a background marker for the image heart.pgm : Fig. 11 c
#    - heart-ObjectMarker.pgm: a pgm image of an object marker for the image heart.pgm : Fig. 11 b
#    - neurite.png: a png image of neurite : Fig. 1 a
#    - neurite-vesselness.png: a png image of the vesselness classification of image neurite.png : Fig. 1 c in the supplemental material
#    - DRIVE/README.txt: link to the DRIVE database and description of the preprocessings done to their images
#    - DRIVE/XX_test_filtered.png with XX from 00 to 20: XXth preprocessed image of the DRIVE database 
#
# Please, download the complete package at http://www.esiee.fr/~perretb/dc-hierarchy.html if one or more files are missing. 
#
# This code is written in Python 3, it requires two free libraries:
# 1 - PIL for image I/O (http://python-imaging.github.io/)
# 2 - pydot for graph plotting in pdf files (https://code.google.com/p/pydot/)
#
# To run this software simply interpret it using a python 3 (eg. execute commande "python DirectedComponentHierarchy.py") 
#
#####################################################################################################
#
# Detailed content:
#
# This code contains 3 variations of the algorithm to construct a directed component hierarchy:
# 1 - "directedComponentHierarchy": Computes the directed component hierarchy of a vertex weighted graph with no optimization
# 2 - "directedComponentHierarchyStackFast": Directed component hierarchy construction of a stack of graphs using Tarjan's union/find
# 3 - "directedComponentHierarchyStack": Directed component hierarchy construction of a stack og graphs with no optimization
#
#
# This code contains 6 demonstration functions:
# 1 - "testAlgoArticle": Construct the hierarchy given in Fig 5 of the article using the algorithm "directedComponentHierarchy"
# 2 - "testAlgoStackGraphArticle": Construct the hierarchy given in Fig 5 of the article using the algorithm "directedComponentHierarchyStack"
# 3 - "testAlgoStackGraphArticleFast": Construct the hierarchy given in Fig 5 of the article using the algorithm "directedComponentHierarchyStackFast"
# 4 - "testNeuriteFiltering": implementation of the neurite image filter (Section 6.2.1)
# 5 - "testHeartSegmentation": implementation of the heart image segmentation (Section 6.2.2)
# 6 - "testRetineAll": implementation of the retinal image filters (Section 6.1)
#
# Contact : benjamin.perret@esiee.fr
#

__author__ = 'Benjamin Perret'

from math import * #@UnusedWildImport

import sys

# for image I/O
try:
    from PIL import Image
    from PIL import ImageTk
except ImportError:
    print("Error: PIL library not available !")
    sys.exit(1)


# for drawing graphs in a pdf file
try:
    import pydot 
    pydotAvailable = True
except ImportError:
    pydotAvailable = False


############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 

### Algorithm of the paper

############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
##############################################################################

# Computes the directed component hierarchy of a vertex weighted graph
# ############################################################################# 
# "Graph" is a directed graph incoded using the internal representation of graphs (see functions "createGraph" and "createEdge")
# "Image" is an array of positive integer weights. Its length must be the number of vertices in "Graph".
# Returns :
#    - the parent relation between the strong connected components of the different upper level sets of the graph
#    - the adjacency relation between the strong connected components of each upper level sets of the graph
def DirectedComponentHierarchy(graph, image):

    # first label of strong components 
    currentLabel=graph['nbPoints']
    # parent relation
    parent=[]
    # adjacency relations
    adj=[]   
    # max weight
    lvlMax=max(image)
    # variables related to the previous level of the hierarchy
    graphPred, labelsPred, currentLabelPred, numLabelsPred = None, None, None, None
    
    # Let's rock !
    for k in range(lvlMax,-1,-1):
        
        # graph induced by vertices of weight >= k
        gk=builgGraph(graph,image,k)
        
        # strong component labelling, labels start at currentLabel
        # lastLabel is the number of the last label used
        labels,lastLabel=strongComponentLabelings(gk, currentLabel)
        # number of labels in the labelling of the level k
        numLabels = lastLabel - currentLabel
        
        # create adjacency relation of the DAG os strong component
        adjk=adjacencyDAG(gk,labels,currentLabel,numLabels)
        adj.extend(adjk)
        
        if k!= lvlMax:
            # create parent relation between current and previous level
            parentk=parentRelation(graphPred, labelsPred, currentLabelPred, numLabelsPred, labels)
            parent.extend(parentk)
        
        # save current level information for future use
        graphPred, labelsPred, currentLabelPred, numLabelsPred = gk, labels, currentLabel, numLabels
        currentLabel = lastLabel
    
    # parents of the last level (if the stack is connected then numLabels==1: the last label is the root of the tree)
    parent.extend([-1]*numLabels)
    
    return parent, adj
    
# Build the graph induced by vertices of "graph" of weight (given by "image") greater than or equal to "level"
# Returns a graph according to the internal representation of this program
def builgGraph(graph, image, level):
    dim=graph['nbPoints']
    
    # create a graph with dim vertices
    # some vertices will certainly be unused.
    # using this trick we are sure that a same vertice get the same index in all levels
    g=createGraph(dim) 
    g["enabled"]=[False]*dim
    
    for i in range(dim):
        # if pixel grey level is greater than or equal to level
        if image[i]>=level:
            # mark vertice as used
            g["enabled"][i]=True
            # get through all edges going out from vertice i in graph
            for j in getAjacentVertices(graph,i):
                if image[j]>=level:
                    createEdge(g,i,j)
    return g
                
# Iterative version of Tarjan's algorithm.
# The directed graph "graph" is labelled into strong components.
# Labels are consecutive integers starting at "firstLabel".
# Returns an array containing the strong component label assigned to each vertex of the graph
def strongComponentLabelings(graph, firstLabel=0):
    
    
    # inner variables
    dim=graph['nbPoints']
    Index=[-1]*dim
    LowLink=[-1]*dim
    IsInStack=[False]*dim

    ind = 0
    stack = []
    derecStack = []
    
    # results
    Labels=[-1]*dim
    curLabel=firstLabel
    
    # recursive sub-procedure of tarjan strong comp construction              
    def strongConnectComp(base):
        nonlocal ind, curLabel
        derecStack.append([base,graph['out'][base]])
        Index[base] = ind
        LowLink[base] = ind
        ind = ind + 1
        IsInStack[base]=True
        stack.append(base)
        while len(derecStack)!=0:
            [i,curAdj]=derecStack.pop() 
            while curAdj!=-1:
                vc = graph['edges'][curAdj][1]
                if Index[vc]==-1 : 
                    derecStack.append([i,curAdj])
                    derecStack.append([vc,graph['out'][vc]])
                    Index[vc] = ind
                    LowLink[vc] = ind
                    ind = ind + 1
                    IsInStack[vc]=True
                    stack.append(vc)
                    break
                elif IsInStack[vc]:
                    LowLink[i] = min(LowLink[i], Index[vc])
                curAdj=graph['nextEdge'][curAdj]

            if curAdj==-1: # all adjacent nodes have been processed
                for vc in getAjacentVertices(graph,i):
                    if Index[vc] > Index[i] and IsInStack[vc]: 
                        LowLink[i] = min(LowLink[i], LowLink[vc])
                    
                # if true i is the "root" of a strong comp, pop the stack until finding i to construct the strong comp
                if(LowLink[i] == Index[i]):
                    w = None 
                    while w != i:
                        w = stack.pop()
                        IsInStack[w]=False
                        Labels[w]=curLabel
                    curLabel=curLabel+1

       
    for i in range(dim):
        if(graph["enabled"][i] and Index[i] == -1):
            strongConnectComp(i)
                     
    return Labels,curLabel

# Construct the adjacency relation on the strong components of the graph "graph"
# "labels" is an array containing the strong component label assigned to each vertex of the graph(see function strongComponentLabelings)
# It is assumed that labels are a serie of "numLalbels" consecutive integers starting at "firstLabel"
# Returns a list of arcs such that [l1 l2] is in  the list iff an arc from l1 to l2 exists in the adjacency relation of strong components
def adjacencyDAG(graph,labels,firstLabel,numLabels):
    dim=graph['nbPoints']
    flags=[-1]*numLabels
    sccs=[]
    for i in range(numLabels):
        sccs.append([])
    adj=[]
    for i in range(dim):
        if graph["enabled"][i]:
            sccs[labels[i]-firstLabel].append(i)
            
    for l in range(numLabels):
        l1 = l + firstLabel
        for x in sccs[l]:
            for y in getAjacentVertices(graph,x):
                l2 = labels[y]
                if l1 != l2 and flags[l2 - firstLabel] != l1:
                    adj.append([l1,l2])
                    flags[l2 - firstLabel] = l1
    return adj
        
# Construct the parent relation between two consecutive levels of the stack
# "GraphPred" is the graph of the previous level
# "LabelsPred" is the strong component labelling of "GraphPred"
# It is assumed that labels in "labelsPred" are a serie of "numLabelsPred" consecutive integers starting at "firstLabelPred"
# "Labels" is the strong component labelling of the vertices of "GraphPred" according to the next graph in the stack
# Returns an array containing for each strong component of label in "labelsPred" the label of its parent strong component.
def parentRelation(graphPred, labelsPred, firstLabelPred, numLabelsPred, labels):
    parent = [-1]*numLabelsPred
    dim=graphPred['nbPoints']
    for i in range(dim):
        if graphPred["enabled"][i]:
            lPar = labels[i]
            l = labelsPred[i]
            parent[l - firstLabelPred] = lPar
    return parent
    
    


############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 

# Graph toolbox : adjacency linked list management for graph "data structure"

############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
##############################################################################

# create a graph with numNodes vertices and no edge
def createGraph(numNodes):
    graph=dict()
    graph['nbPoints']=numNodes
    graph['out']=[-1]*numNodes
    graph['outTail']=[-1]*numNodes
    # an edge is an array of two elements [source,destination] so this is kind of a hybrid graph representation
    graph['edges']=[]
    graph['prevEdge']=[]
    graph['nextEdge']=[]
    return graph

# create a new edge in the graph
def createEdge(graph, source, dest, weight=None):
    i=len(graph['prevEdge'])
    if weight!=None:
        graph['edges'].append([source,dest,weight])
    else:
        graph['edges'].append([source,dest])
    graph['prevEdge'].append(-1)
    graph['nextEdge'].append(-1)
    setEdge(graph, source, dest, i)

# get a list of vertices adjacent to the given node
def getAjacentVertices(graph, node):
    nodes = []
    e = graph['out'][node]
    while(e != -1):
        # dest is the adjancent vertex
        nodes.append(graph['edges'][e][1])
        # next edge in the edge list
        e = graph['nextEdge'][e]
    return nodes

# set the edge i of the graph, the edge is put at the head of the out linked list of the source vertex
def setEdge(graph, source, dest, i):
    n1 = source
    if graph['out'][n1] == -1:
        graph['outTail'][n1] = i
    else:
        e1 = graph['out'][n1]
        graph['prevEdge'][e1] = i
        graph['nextEdge'][i] = e1
    graph['out'][n1] = i

# concatenate edge lists, out edges of vertex n2 are added to vertex n1
def concatEdgeOut(graph, n1, n2):
    if graph['outTail'][n1] != -1:
        graph['nextEdge'][graph['outTail'][n1]] = graph['out'][n2]
    if graph['out'][n2] != -1:    
        graph['prevEdge'][graph['out'][n2]] = graph['outTail'][n1]
    if graph['outTail'][n2] != -1:
        graph['outTail'][n1] = graph['outTail'][n2]
    graph['out'][n2] = graph['outTail'][n2] = -1

# remove the edge e of the list of out edges of vertex n
def removeEdgeOut(graph, n, e):
    if graph['out'][n] == e:
        graph['out'][n] = graph['nextEdge'][e]
    if graph['outTail'][n] == e:
        graph['outTail'][n] = graph['prevEdge'][e]
    if graph['nextEdge'][e] != -1:
        graph['prevEdge'][graph['nextEdge'][e]] = graph['prevEdge'][e]
    if graph['prevEdge'][e] != -1:
        graph['nextEdge'][graph['prevEdge'][e]] = graph['nextEdge'][e]
    graph['nextEdge'][e] = graph['prevEdge'][e] = -1


############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 

### Semi-naive algorithm for directed component hierarchy of a stack
###
### Contracts the graph into its DAG of strong connected components 
### at each level and uses the compressed graph as the basis of the
### next level.

############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 

# Union find 
############################################################################## 
# find canonical node of elem with path compression
def findTarjan(elem, Par):
    i = elem
    while Par[i] >= 0:
        i = Par[i]
    while Par[elem] >= 0:
        temp = elem
        elem = Par[elem]
        Par[temp] = i
    return i

# union of nodes i and j with ranking, the chosen canonical node is returned
# and the successors of the other nodes are transferred to the canonical one
def unionTarjan(graph, i, j, Par, Rank):
    if Rank[i] > Rank[j]:
        i, j = j, i
    elif Rank[i] == Rank[j]:
        Rank[j] = Rank[j] + 1
    Par[i] = j
    concatEdgeOut(graph, j, i)  # perhaps we can be more fancy here
    return j




# Directed component hierarchy construction from a stack using tarjan union find
############################################################################## 
# "stack" stack of graphs : see functions createGraphStackFromEdgeWeightedGraph and createGraphStackFromVertexWeightedGraph
# "completeGraph" if False then only the strong component tree is built.
#                 if True then the whole directed component hierarchy is built.
# Returns :
#    - the parent relation between the strong connected components of the different upper level sets of the graph
#    - the adjacency relation between the strong connected components of each upper level sets of the graph

def directedComponentHierarchyStackFast(stack, completeGraph=True):
    completeGraphEdges = []
    nbPoints = stack['nbPoints']
    
    # pre-allocations of lots of things
    
    # union find makeset
    Par = [-1] * nbPoints
    parent = [-1] * nbPoints
    Root = [i for i in range(nbPoints)]
    Rank = [0] * nbPoints
    Lvls = [-1] * nbPoints
    # strong comp detection
    Index = [-1] * nbPoints
    LowLink = [-1] * nbPoints
    IsInStack = [False] * nbPoints
    # list of canonical vertices in union find
    activeVertices = []  # nodes
    edges = stack['edges']
    i = 0
    c = 0
    while i < len(edges):
        lvl = stack['lvlInsertionSequence'][c]
        # add nodes at lvl c
        newnodes = stack['pointInsertionSequence'][c]
        c += 1
        for n in newnodes:
            Lvls[n] = lvl
        activeVertices.extend(newnodes)
    
        # add all edges at level lvl (assuming edges are sorted)              
        while i < len(edges) and edges[i][2] == lvl:
            u = edges[i][0]
            v = edges[i][1]
            uc = findTarjan(u, Par)
            vc = findTarjan(v, Par)
            if(uc!=vc):
                setEdge(stack, uc, vc, i)
            i = i + 1
        activeVertices = strongConnectDerecFast(stack, activeVertices, Par, Rank, Root, parent, Index, LowLink, IsInStack, completeGraph, Lvls, lvl)
        if completeGraph:
            saveCompleteGraphEdges(stack, activeVertices, Par, Root, completeGraphEdges)
    return parent, completeGraphEdges, Lvls


# performs a contraction into a DAG of strong connected comps and maintain the strong comp tree,
# returns all the list of canonical nodes representing the strong connected components
def strongConnectDerecFast(graph, activeVertices, Par, Rank, Root, parent, Index, LowLink, IsInStack, completeGraph, Lvls, lvl):
    ind = 0
    stack = []
    derecStack = []
    NActiveVertices = []
    for i in activeVertices:
        Index[i] = -1
        IsInStack[i] = False
        graph['flags'][i] = -1
    
    # recursive sub-procedure of tarjan strong comp construction              
    def strongConnectComp(base):
        nonlocal ind, completeGraph
        derecStack.append([base,graph['out'][base]])

        Index[base] = ind
        LowLink[base] = ind
        IsInStack[base]=True
        stack.append(base)
        ind = ind + 1

        while len(derecStack)!=0:
            [i,curAdj]=derecStack.pop()

            while curAdj != -1:
                vc = findTarjan(graph['edges'][curAdj][1], Par)
                if Index[vc]==-1: # first touch for this adjacent vertex
                    derecStack.append([i,curAdj])
                    derecStack.append([vc,graph['out'][vc]])
                    Index[vc] = ind
                    LowLink[vc] = ind
                    ind = ind + 1
                    IsInStack[vc]=True
                    stack.append(vc)
                    break
                elif  IsInStack[vc]: # this node has already been touched by someone else and is not in another scc
                    LowLink[i] = min(LowLink[i], Index[vc])
                
                curAdj=graph['nextEdge'][curAdj]
    
            if curAdj==-1: # all adjacent nodes have been processed
                e = graph['out'][i]
                # compute lowlink
                while e != -1:
                    vc = findTarjan(graph['edges'][e][1], Par)
                    # adjacent node that were not already processed during first pass and that have not been added in another scc
                    if Index[vc] > Index[i] and IsInStack[vc]: 
                        LowLink[i] = min(LowLink[i], LowLink[vc])

                    e = graph['nextEdge'][e]
                
                # if true i is the "root" of a strong comp, pop the stack until finding i to construct the strong comp
                if(LowLink[i] == Index[i]):
                    tmp = i
                    w = stack.pop()
                    IsInStack[w]=False
                    if(w != i) or completeGraph:  # else scc has not changed
                        newNode = len(parent)
                        parent.append(-1)
                        Lvls.append(lvl)
                        Root.append(newNode)
                        parent[Root[i]] = newNode
                        while w != i:
                            tmp = unionTarjan(graph, tmp, w, Par, Rank)
                            parent[Root[w]] = newNode
                            w = stack.pop()
                            IsInStack[w]=False
                        Root[tmp] = newNode
                        cleanOutEdges(graph, tmp, Par)       
                    NActiveVertices.append(tmp)  # tmp is the canonical node, representative of the strong comp in the DAG

       
    for i in activeVertices:
        if(Index[i] == -1):
            strongConnectComp(i)
                     
    return NActiveVertices




# remove reflexive and redundant out edges from vertice i                     
def cleanOutEdges(graph, i, Par):
    e = graph['out'][i]
    while e != -1:
        nexte = graph['nextEdge'][e]
        vc = findTarjan(graph['edges'][e][1], Par)
        if vc == i or graph['flags'][vc] == i:
            removeEdgeOut(graph, i, e)
        else:
            graph['flags'][vc] = i
        e = nexte


# save the edges of the dag of strong connected components at the current level in the list completeGraphEdges
def saveCompleteGraphEdges(graph, activeVertices, Par, Root, completeGraphEdges):
    for i in activeVertices:
        source = Root[i]
        e = graph['out'][i]
        while e != -1:
            vc = findTarjan(graph['edges'][e][1], Par)
            dest = Root[vc]
            completeGraphEdges.append([source, dest])
            e = graph['nextEdge'][e]


############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 

### Completely Naive algorithm for any stack of graphs

############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 


# Directed component hierarchy construction from a stack with no optimization
############################################################################## 
# "stack" stack of graphs : see functions createGraphStackFromEdgeWeightedGraph and createGraphStackFromVertexWeightedGraph
# Returns :
#    - the parent relation between the strong connected components of the different upper level sets of the graph
#    - the adjacency relation between the strong connected components of each upper level sets of the graph
#    - the number of arcs going out of each strong connected component

def directedComponentHierarchyStack(stack):
    nbPoints = stack['nbPoints']
    activeVertices = []  # nodes
    parent = [-1] * nbPoints
    Lvls = [-1] * nbPoints
    nbOut =[0] * nbPoints
    
    Index = [-1] * nbPoints
    LowLink = [-1] * nbPoints
    IsInStack = [0] * nbPoints
    strongCompLvL = [i for i in range(nbPoints)]
    strongCompLvLPrec = [i for i in range(nbPoints)]

    dagEdges =[]
    
    edges = stack['edges']
    i = 0
    c = 0

    while i < len(edges):
        lvl = stack['lvlInsertionSequence'][c]
        # add nodes at lvl c
        newnodes = stack['pointInsertionSequence'][c]
        c += 1
        for n in newnodes:
            Lvls[n] = lvl
        activeVertices.extend(newnodes)

        # add all edges at level lvl (assuming edges are sorted)
              
        while i < len(edges) and edges[i][2] == lvl:
            u = edges[i][0]
            v = edges[i][1]
            setEdge(stack, u, v, i)
            i = i + 1
        
        strongConnectDerec(stack,activeVertices, Index, LowLink, IsInStack, strongCompLvL,dagEdges,parent,Lvls,lvl,nbOut)
        
        
        for j in range(nbPoints):
            parent[strongCompLvLPrec[j]]=strongCompLvL[j]
        
        strongCompLvLPrec, strongCompLvL =  strongCompLvL, strongCompLvLPrec         

       
    return parent, dagEdges, Lvls, nbOut 


def strongConnectDerec(graph,activeVertices, Index, LowLink, IsInStack, strongCompLvL,dagEdges,parent,Lvls,lvl,nbOut):
    ind = 0
    stack = []
    derecStack = []
    
    for i in activeVertices:
        Index[i] = -1
        IsInStack[i] = False  
        graph['flags'][i] = -1
                 
    def strongConnectComp(base):
        nonlocal ind
        derecStack.append([base,graph['out'][base]])

        Index[base] = ind
        LowLink[base] = ind
        IsInStack[base]=True
        stack.append(base)
        ind = ind + 1

        def addOutEdges(lst, n):
            oe = graph['out'][n]
            while oe != -1:
                v = graph['edges'][oe][1]
                lst.append(v)
                oe = graph['nextEdge'][oe]
        
        def cleanOutEdgesNaif(i, out):
            marked=[True]*len(out)
            for j in range(len(out)):
                vc = strongCompLvL[out[j]]
                if vc == i:
                    marked[j]=False
                elif graph['flags'][vc] == i:
                    nbOut[i] = nbOut[i] +1
                    marked[j]=False
                else:
                    nbOut[i] = nbOut[i] +1
                    graph['flags'][vc] = i
            nout = [strongCompLvL[out[j]] for j in range(len(out)) if marked[j] ]
            return nout

        while len(derecStack)!=0:
            [i, curAdj]=derecStack.pop()

                     
            while curAdj != -1:
                vc = graph['edges'][curAdj][1]
                if Index[vc]==-1: # first touch for this adjacent vertex
                    derecStack.append([i,curAdj])
                    derecStack.append([vc,graph['out'][vc]])
                    Index[vc] = ind
                    LowLink[vc] = ind
                    ind = ind + 1
                    IsInStack[vc]=True
                    stack.append(vc)
                    break

                elif  IsInStack[vc]: # this node has already been touched by someone else and is not in another scc
                    LowLink[i] = min(LowLink[i], Index[vc])
                curAdj=graph['nextEdge'][curAdj]

            if curAdj==-1: # all adjacent nodes have been processed

                e = graph['out'][i]
                # compute lowlink
                while e != -1:
                    vc = graph['edges'][e][1]
                    # adjacent node that were not already processed during first pass and that have not been added in another scc
                    if Index[vc] > Index[i] and IsInStack[vc]: 
                        LowLink[i] = min(LowLink[i], LowLink[vc])
                    e = graph['nextEdge'][e]
                
                # if true i is the "root" of a strong comp, pop the stack until finding i to construct the strong comp
                if(LowLink[i] == Index[i]):
                    w = None
                    comp=len(parent)
                    parent.append(-1)
                    graph['flags'].append(-1)
                    strongCompLvL[i]=comp
                    Lvls.append(lvl)
                    nbOut.append(0)
                    out = []
           
                    while w != i:
                        w = stack.pop()
                        strongCompLvL[w]=comp
                        IsInStack[w]=False
                        addOutEdges(out,w)
                
                    out = cleanOutEdgesNaif(comp, out)
            
                    for i in out:
                        dagEdges.append([comp, i])
       
    for i in activeVertices:
        if(Index[i] == -1):
            strongConnectComp(i)
            
############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 

### Graph reading, image reading and stack construction (differential coding) 

############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 


# Create a 4 or 8-neighborhood graph for an image of dimensions "size"
# k=4 or 8 indicates the adjacency.
# If "graph" equals None a new graph is created, otherwise the edges are added to the existing graph.
def createSimpleNeighbourhoodGraph(size,k=4, graph=None):
    if graph==None:
        graph =  createGraph(size[0]*size[1])
    width=size[0]
    height=size[1]
    coordLinTo2D = lambda x: (x % width, x // width)
    coord2DToLin= lambda x,y: y*width+x
    dim = width * height

    def writeLink(i,j):
        if not j in getAjacentVertices(graph, i):
            createEdge(graph,i,j)
         
    for i in range(dim):
        x,y = coordLinTo2D(i)
        if(x+1<width):
            writeLink(i,coord2DToLin(x+1,y))
        if(x-1>=0):
            writeLink(i,coord2DToLin(x-1,y))
        if(y+1<height):
            writeLink(i,coord2DToLin(x,y+1))
        if(y-1>=0):
            writeLink(i,coord2DToLin(x,y-1))
        if(k==8):
            if(x+1<width and y-1>=0):
                writeLink(i,coord2DToLin(x+1,y-1))
            if(x-1>=0 and y-1>=0):
                writeLink(i,coord2DToLin(x-1,y-1))
            if(x+1<width and y+1<height):
                writeLink(i,coord2DToLin(x+1,y+1))
            if(x-1>=0 and y+1<height):
                writeLink(i,coord2DToLin(x-1,y+1))        
    
    return graph


# Create an edge weighted symmetric directed graph.
# The graph nodes are thee image pixels.
# Arcs are symmetric and constructed using a 4 or 8 connectivity.
# Weights are computed using the function "weightFunction" which is not necessarily symmetric.
# Weight function must a be a function that associates a dissimilarity measure to two grey level values 
# (the grey level value of the origin of the arc and the one of its destination)
def createHeartGradientGraph(image, size, weightFunction, connectivity=8):
    graph =  createGraph(size[0]*size[1])
    width=size[0]
    height=size[1]
    coordLinTo2D = lambda x: (x % width, x // width)
    coord2DToLin= lambda x,y: y*width+x
    dim = width * height
    
    def writeLink(i,j):
        createEdge(graph,i,j,weightFunction(image[i],image[j]))
        
    for i in range(dim):
        x,y = coordLinTo2D(i)
        if(x+1<width):
            writeLink(i,coord2DToLin(x+1,y))
        if(x-1>=0):
            writeLink(i,coord2DToLin(x-1,y))
        if(y+1<height):
            writeLink(i,coord2DToLin(x,y+1))
        if(y-1>=0):
            writeLink(i,coord2DToLin(x,y-1))
        if(connectivity==8):
            if(x+1<width and y-1>=0):
                writeLink(i,coord2DToLin(x+1,y-1))
            if(x-1>=0 and y-1>=0):
                writeLink(i,coord2DToLin(x-1,y-1))
            if(x+1<width and y+1<height):
                writeLink(i,coord2DToLin(x+1,y+1))
            if(x-1>=0 and y+1<height):
                writeLink(i,coord2DToLin(x-1,y+1))        
    
    return graph
    
# Create the graph for the neurite application.
# Image is the vesselness image.
def createSemanticGraphNeurite(image,size, tube=1,bulbe=2):
    graph =  createGraph(size[0]*size[1])
    width=size[0]
    height=size[1]
    coordLinTo2D = lambda x: (x % width, x // width)
    coord2DToLin= lambda x,y: y*width+x
    dim = width * height

    def writeLink(i,j):
        if image[i]==image[j]:
            createEdge(graph,i,j)
        elif  (image[i]==tube and image[j]==bulbe):
            createEdge(graph,i,j)
        elif  image[i]==0:
            createEdge(graph,i,j)

            
    for i in range(dim):
        x,y = coordLinTo2D(i)
        if(x+1<width):
            writeLink(i,coord2DToLin(x+1,y))
        if(x-1>=0):
            writeLink(i,coord2DToLin(x-1,y))
        if(y+1<height):
            writeLink(i,coord2DToLin(x,y+1))
        if(y-1>=0):
            writeLink(i,coord2DToLin(x,y-1))
        if(x+1<width and y-1>=0):
            writeLink(i,coord2DToLin(x+1,y-1))
        if(x-1>=0 and y-1>=0):
            writeLink(i,coord2DToLin(x-1,y-1))
        if(x+1<width and y+1<height):
            writeLink(i,coord2DToLin(x+1,y+1))
        if(x-1>=0 and y+1<height):
            writeLink(i,coord2DToLin(x-1,y+1))        
    
    return graph

# Symmetrize the graph with the max strategy: whenever an edge (p,q) is found the edge (q,p) is added to the graph
def symmetrizeGraphMax(graph):
    g =  createGraph(graph["nbPoints"])
    for i in range(graph["nbPoints"]):
        for j in getAjacentVertices(graph, i):
            if not j in getAjacentVertices(g, i):
                createEdge(g,i,j)
            if not i in getAjacentVertices(g, j):
                createEdge(g,j,i)
    return g

# Symmetrize the graph with the min strategy: an edge (p,q) is preserved only if the edge (q,p) is also in the graph
def symmetrizeGraphMin(graph):
    g =  createGraph(graph["nbPoints"])
    for i in range(graph["nbPoints"]):
        for j in getAjacentVertices(graph, i):
            if j>i and i in getAjacentVertices(graph, j):
                    createEdge(g,i,j)
                    createEdge(g,j,i)
    return g

# transpose the graph: each edge (p,q) is transformed into the edge (q,p)
def transposeGraph(graph):
    g =  createGraph(graph["nbPoints"])
    for i in range(graph["nbPoints"]):
        for j in getAjacentVertices(graph, i):
            createEdge(g,j,i)
    return g



# Read an image and put it in a 1D array
def readImage(imageFile):
    image= Image.open(imageFile)
    size = image.size
    width = size[0]
    height = size[1]
    coord = lambda x: (x % width, x // width)
    dim = width * height
    pixdata = image.load()
    
    levels = []
    for i in range(dim):
        levels.append(pixdata[coord(i)])
    
    return levels, size

# Read a graph from ascii file.
# format (weights are optionnale, vertices are numbered from 0 to numberOfVertices-1):
# numberOfVertices numberOfEdges
# sourceVertexOfEdge_1 destinationVertexOfEdge_1 weightOfEdge_1
# sourceVertexOfEdge_2 destinationVertexOfEdge_2 weightOfEdge_2
# ...
# sourceVertexOfEdge_NumberOfEdges destinationVertexOfEdge_NumberOfEdges weightOfEdge_NumberOfEdges
def readGraph(filename):
    infile = open(filename, "r")
    line = infile.readline()
    while(line[0] == '#'):
        line = infile.readline()
    nbPoints = int(line.split()[0])
    lines = infile.readlines()
    infile.close()

    graph=createGraph(nbPoints)
    for line in lines:
        tokens = line.split()
        if len(tokens)==3:
            createEdge(graph,int(tokens[0]), int(tokens[1]), int(tokens[2]))
        else:
            createEdge(graph,int(tokens[0]), int(tokens[1]))
    
    return graph

# Save a graph in a file as plain text
def saveGraph(graph, filename):
    out = open(filename, "w")
    out.write(str(graph["nbPoints"]) + " " + str(len(graph["edges"])) + "\n")
    for e in graph["edges"]:
        line=" ".join([str(v) for v in e])
        out.write(line)
        out.write("\n")
    out.close()

# Create a stack from the edge weighted graph 
def createGraphStackFromEdgeWeightedGraph(graph):
    dim=graph["nbPoints"]
    stack = createGraph(dim)
    edges = graph["edges"]
    nbe = len(edges)
    #stack["edges"].extend(edges)
    
    el = [i for i in range(nbe)]
    el.sort(key=lambda x:edges[x][2])
    
    added = [False]*dim
    addedNodes = []
    addedLvls =[]
    stack['pointInsertionSequence'] = addedNodes
    stack['lvlInsertionSequence'] = addedLvls
    stackEdges = stack['edges']
    
    c=0
    
    while c<len(edges):
        e=edges[el[c]]
        lvl=e[2]
        addedLvls.append(lvl)
        nnodes = []
        while(c<len(edges) and edges[el[c]][2]==lvl):
            e=edges[el[c]]
            stackEdges.append(e)
            if not added[e[0]]:
                nnodes.append(e[0])
                added[e[0]]=True
            if not added[e[1]]:
                nnodes.append(e[1])
                added[e[1]]=True
            c = c + 1
        addedNodes.append(nnodes)
    
    stack['nextEdge'] = [-1] * len(edges)
    stack['prevEdge'] = [-1] * len(edges)
    stack['flags'] = [-1] * dim
    return stack

# Construct a stack for a vertex weighted graph.   
# Weights are given by the image.
def createGraphStackFromVertexWeightedGraph(graph,image):
    dim = len(image)
    stack = createGraph(dim)
    
    pix = [i for i in range(dim)]
 
    pix.sort(key=lambda x:image[x]) 

    addedNodes = []
    addedLvls =[]
    stack['pointInsertionSequence'] = addedNodes
    stack['lvlInsertionSequence'] = addedLvls
    edges = stack['edges']
   
    for e in graph["edges"]:
        x=e[0]
        y=e[1]
        edges.append([x, y, min(image[x],image[y])])
        

    i = dim - 1
    while i >= 0:
        j = pix[i]
        lvl = image[j]
        
        nnodes = []
        while(i >= 0 and image[j] == lvl):
            nnodes.append(j)
            i -= 1
            j = pix[i]
        addedLvls.append(lvl)
        addedNodes.append(nnodes)

    edges.sort(key=lambda x:-x[2])  # lambda x,y: cmp(x[2],y[2]) )

    stack['nextEdge'] = [-1] * len(edges)
    stack['prevEdge'] = [-1] * len(edges)
    stack['flags'] = [-1] * dim
    return stack


############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 

### Usefull functions to finalyze the structure

############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 

# takes a parent relation and the Lvls of each node. 
# If the parent relation does not define a tree, a root is created 
# and each tree of the forest is rooted in this new node.
# The new root is assigned to level -1
def ensureTree(parent, Lvls=[]):
    if parent.count(-1) > 1:
        nel = len(parent)
        for i in range(len(parent)):
            if(parent[i] == -1):
                parent[i] = nel
        parent.append(-1)
        Lvls.append(-1)  # yeark

# Construct a more usable structure from the result of the algorithm directedComponentHierarchyXXXXXX
def buildFinalGraph(nbPoints, parent, completeGraphEdges, Lvls, nbOut=[]):
    graph = dict()
    graph["nbPoints"] = nbPoints
    graph["leaves"] = range(nbPoints)
    nodes = []
    graph["nodes"] = nodes
    # allocate nodes
    for i in range(len(parent)):
        node = dict()
        node["index"] = i
        node["level"] = Lvls[i]
        node["parent"] = parent[i]
        node["sucs"] = []
        node["preds"] = []
        node["children"] = []
        if len(nbOut)>i:
            node["nbout"] = nbOut[i]
        nodes.append(node)
    # define children relation
    for i in range(len(parent)):
        par = parent[i]
        if(par != -1):
            nodes[par]["children"].append(i)
    # define siblings
    for e in completeGraphEdges:
        nodes[e[0]]["sucs"].append(e[1])
        nodes[e[1]]["preds"].append(e[0])
    return graph




############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 

### Attributes

############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 

# Compute the level of each node as its depth in the tree given by parent relation
def computeLevelsAsDepth(parent):
    levels = [0]*len(parent)
    for j in range(len(levels)-1,-1,-1):
        if parent[j]!=-1:
            levels[j] = levels[parent[j]]+1
    return levels

# Compute the attribute "attributeName" on the graph.
# For each node, the attribute is set to true if the strong connected 
# component of the node contains a point of the image "markerImage" having the value "markerValue"
def computeAttributeMarker(graph, markerImage, markerValue, attributeName):
    nbPoints = graph["nbPoints"]
    
    nodes = graph["nodes"]

    for i in range(nbPoints):
        nodes[i][attributeName]= (markerImage[i]==markerValue)
    
    for i in range(nbPoints,len(nodes),1):
        flag=False
        for j in nodes[i]["children"]:
            if nodes[j][attributeName]:
                flag = True
                break
        nodes[i][attributeName]=flag
            
# Compute the area (number of nodes) of each SCC (attribute "area") and each DCC (attribute "area_semi")                   
def computeAttributeArea(graph):
    nodes = graph["nodes"]
    nbPoints = graph["nbPoints"]
    for i in range(len(nodes)):
        if(i < nbPoints):
            nodes[i]["area"] = 1
            nodes[i]["area_semi"] = None
        else:
            nodes[i]["area"] = 0
            nodes[i]["area_semi"] = None
    for i in range(len(nodes)):
        par = nodes[i]["parent"]
        if(par != -1):
            nodes[par]["area"] += nodes[i]["area"]

    visit=[None]*len(nodes)

    def computeSemi(nodes,i,ref):
        visit[i]=ref
        tmp=nodes[i]["area"]
        for j in nodes[i]["sucs"]:
            if(visit[j]!=ref):
                tmp=tmp+computeSemi(nodes,j,ref)
        return tmp
                
    for i in range(len(nodes)):
        nodes[i]["area_semi"]=computeSemi(nodes,i,i)

# Compute the coordinate (position in the image grid) of each leave of hierarchy (attribute "coords")        
def computeAttributeLeavesCoordinates(graph, width):
    nodes = graph["nodes"]
    for i in graph["leaves"]:
        nodes[i]["coords"] = (i % width, i // width)
    for i in range(len(graph["leaves"]),len(nodes)):
        nodes[i]["coords"] =  nodes[nodes[i]["children"][0]]["coords"]

# Compute the moments [M00 M10 M01 M11 M20 M02] of each SCC (attribute "moments") and each DCC  (attribute "moments_semi")
def computeAttributeSimpleMoments(graph):
    nodes = graph["nodes"]
    nbPoints = graph["nbPoints"]
    for i in range(len(nodes)):
        if(i < nbPoints):
            c = nodes[i]["coords"]
            x = c[0]
            y = c[1]
            nodes[i]["moments"] = [1, x, y, x * y, x * x, y * y]
            nodes[i]["moments_semi"] = [1, x, y, x * y, x * x, y * y]
        else:
            nodes[i]["moments"] = [0, 0, 0, 0, 0, 0]
            nodes[i]["moments_semi"] = None
    for i in range(len(nodes)):
        par = nodes[i]["parent"]
        if(par != -1):
            m = nodes[i]["moments"]
            mp = nodes[par]["moments"]
            for j in range(len(m)):
                mp[j] += m[j]

    visit=[None]*len(nodes)

    def computeSemi(nodes,i,ref):
        visit[i]=ref
        tmp=nodes[i]["moments"]
        for j in nodes[i]["sucs"]:
            if(visit[j]!=ref):
                y=computeSemi(nodes,j,ref)
                tmp=[ tmp[i] + y[i] for i in range(len(tmp))]
        return tmp
                
    for i in range(len(nodes)):
        nodes[i]["moments_semi"]=computeSemi(nodes,i,i)

# Compute the moment of inertia of each SCC (attribute "inertia") and each DCC  (attribute "inertia_semi")
def computeAttributeInertia(graph):
    def compInit(m):
        xmean = m[1] / m[0]
        ymean = m[2] / m[0]
        u20 = m[4]-xmean*m[1]
        u02 = m[5]-ymean*m[2]
        return (u20+u02)/(m[0]*m[0])
    
    nodes = graph["nodes"]
    for i in range(len(nodes)):
        nodes[i]["inertia"]=compInit(nodes[i]["moments"])
        nodes[i]["inertia_semi"]=compInit(nodes[i]["moments_semi"])
      
        
# Compute the elongation and the orientation from an array of moments [M00 M10 M01 M11 M20 M02]    
def elongationOrientationFromMoments(m):
    sign = lambda x: copysign(1, x)
    xmean = m[1] / m[0]
    ymean = m[2] / m[0]
    xvar = m[4] / m[0] - xmean * xmean
    yvar = m[5] / m[0] - ymean * ymean
    xycovar = m[3] / m[0] - xmean * ymean
    if(xvar - yvar != 0):
        a = 0.5 * atan(2 * xycovar / (xvar - yvar))
        if(sign(a) * sign(xycovar) < 0.0):
            a += pi / 2.0;
        alpha = a
    else:
        if xycovar == 0:
            alpha = 0
        else:
            alpha = pi / 4.0 
    lambda1 = max(0,0.5 * (xvar + yvar + sqrt(4 * xycovar * xycovar + (xvar - yvar) * (xvar - yvar))))
    lambda2 = max(0,0.5 * (xvar + yvar - sqrt(4 * xycovar * xycovar + (xvar - yvar) * (xvar - yvar))))

    
    if ( lambda1 == 0): # ill posed case
        el = 1;
    elif(lambda2==0): # ill posed case
        el = sqrt((lambda2+1) / (lambda1+1));
    else :
        el = sqrt(lambda2 / lambda1);

    return el, alpha
                     
# Compute the elongation and orientation of each SCC (attributes "elongation" and "orientation") and each DCC  (attribute "elongation_semi" and "orientation_semi")
def computeAttributeElongationOrientation(graph):
    nodes = graph["nodes"]
    for i in range(len(nodes)):
        nodes[i]["elongation"], nodes[i]["orientation"] = elongationOrientationFromMoments(nodes[i]["moments"])
        nodes[i]["elongation_semi"], nodes[i]["orientation_semi"] = elongationOrientationFromMoments(nodes[i]["moments_semi"])

    


# Compute the attribute value "attributeName_directed" on the graph.
# For each node, the attribute is set to true if the directed connected 
# component rooted int it contains a node having its attribute "attributeName" set to true.
# see function computeAttributeMarker
def computeAttributeMarkerDirected(graph,attributeName):
    attributeNameSemi = attributeName +"_directed"
    
    nodes = graph["nodes"]
    for i in range(len(nodes)):
        nodes[i][attributeNameSemi]=False
    
    def propagateSemiPred(i):
        for p in nodes[i]["preds"]:
            if nodes[p][attributeNameSemi]==False:
                nodes[p][attributeNameSemi]=True
                propagateSemiPred(p)
               
            
    for i in range(len(nodes)): 
        if nodes[i][attributeName] and not nodes[i][attributeNameSemi]:
            nodes[i][attributeNameSemi]=True
            propagateSemiPred(i)


# draw directed component hierarchy in a pdf with the values of the specified attributes
# "name" : filename of the produced pdf
# "graph" : graph to plot
# "attributes" : list of attributes values to print
# "leaves" : if true leaves are ploted
def drawGraphVizAttr(name, graph, attributes=[], leaves=False):
    if not pydotAvailable:
        print ("Error: this function requires the library pydot (which requires the software graphviz.")
        return
    
    #n = len(parent)
    G = pydot.Dot(graph_type='graph')

    nodes = graph["nodes"]
    nbPoints = graph["nbPoints"]
    for i in range(len(nodes)):
        if(i>=nbPoints or leaves):
            node = pydot.Node(i)
            if(i < nbPoints):
                node.set_shape("square")
            label = str(i)
            if attributes != []:
                labels = ",".join([str(nodes[i][a]) for a in attributes])
                label += "(" + labels + ")"
            node.set_label(label)
            G.add_node(node)
       
    for i in range(len(nodes)):
        if(i>=nbPoints or leaves):
            par = nodes[i]["parent"]
            if(par != -1):
                edge = pydot.Edge(par, i)
                edge.set_dir("forward")
                G.add_edge(edge)
    
            for d in nodes[i]["sucs"]:
                edge = pydot.Edge(i, d)
                edge.set_dir("forward")
                edge.set_style("dotted")
                edge.set_constraint("false")
                G.add_edge(edge)
       

    G.write_pdf(name)


    
############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 

### Node selection

############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 

# test if a given node is marked as deleted
def isDeleted(node):
    if not("deleted" in node):
        return False
    return node["deleted"]

# apply a given selection criterion f on the graph
# the selection criterion is a function that associates the value true or false to a node
def filterDirect(graph, f):
    nodes = graph["nodes"]
    for i in range(len(nodes)):
        nodes[i]["deleted"] = None

    for i in range(len(nodes)):
        nodes[i]["deleted"] = f(graph, nodes[i])

# regularize the result of a selection criterion using the select max strategy
def regularizeSelectMax(graph):
    nodes = graph["nodes"]

    def keepDCC(i):
        nodes[i]["deleted"] = False
        for c in nodes[i]["sucs"]:
            if nodes[c]["deleted"]:
                keepDCC(c)

    for i in range(len(nodes)):
        if not nodes[i]["deleted"]:   
            keepDCC(i)


# regularize the result of a selection criterion using the select max hierarchical strategy               
def regularizeSelectMaxHierarchical(graph):
    nodes = graph["nodes"]

    def keepBranch(i):
        nodes[i]["deleted"] = False
        for c in nodes[i]["children"]:
            if nodes[c]["deleted"]:
                keepBranch(c)

    for i in range(len(nodes)):
        if not nodes[i]["deleted"]:   
            keepBranch(i)

# regularize the result of a selection criterion using the discard min strategy              
def regularizeDiscardMin(graph):
    nodes = graph["nodes"]

    def removeDCC(i):
        nodes[i]["deleted"] = True
        for c in nodes[i]["sucs"]:
            removeDCC(c)

    for i in range(len(nodes)):
        if nodes[i]["deleted"]:   
            removeDCC(i)

############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 

### Filtering

############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
  
# Compute the reconstruction value of every leaf of the graph.
# The reconstruction value of a leaf is the attribute value of the attribute "attr"
# of its closest non deleted ancestor.
# If all the ancestors of a leaf are deleted its reconstruction value is set to "background"
def reconstructValue(graph, attr, background=0):
    nodes = graph["nodes"]
    
    for i in range(len(nodes)-1,graph["nbPoints"]-1,-1):
        if isDeleted(nodes[i]):
            p = nodes[i]["parent"]
            if p!=-1:
                v = nodes[p]["reconstruction"]
            else:
                v = background  # oops
        else:
            v = nodes[i][attr]
        nodes[i]["reconstruction"]=v
    
    for i in range(graph["nbPoints"]):
        p = nodes[i]["parent"]
        nodes[i]["reconstruction"] = nodes[p]["reconstruction"]              

# Save a reconstruction of the graph in an image, i.e. puts the reconstruction value of the leaves in an image
# see function reconstructValue for more details
# "graph" the graph to reconstruct
# "attr" the attribute that will give the gray level values of the image
# "name" filename of the saved image
# "width" width of the image (height is deduced from the number of leaves in the graph)
# "background" background gray level value  
def saveAsImage(graph, attr, name, width,background=0):
    im=getAsImage(graph, attr, width,background)
    im.save(name, "PNG")
    
def getAsImage(graph, attr, width,background=0):
    reconstructValue(graph, attr, background)
    nodes = graph["nodes"]
    size = width, graph["nbPoints"] // width
    im = Image.new("L", size, 0)
    pix = im.load()
    for i in range(graph["nbPoints"]):
        v = nodes[i]["reconstruction"]
        if type(v) == bool:
            if v:
                v= 255
            else:
                v=0
        pix[i % width, i // width] = int(v)
    return im 
    


############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 

### Tests

############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
##############################################################################


# Test of the algorithm "as close as possible" of the one given in the article 
# Input: 
#    - file fig5-Graph.list : plain text file encoding the graph of Fig. 5
#    - file fig5-Image.png: png image encoding the weights of the graph of Fig. 5
# Output:
#    - file "DCC-Hierarchy-Fig5.pdf": pdf plot of the graph Fig. 6 c
def testAlgoArticle():
    print("Test of the paper algorithm for vertex weighted graphs")
    graph = readGraph("fig5-Graph.list")
    image, size=readImage("fig5-Image.png")
    parent, adj= DirectedComponentHierarchy(graph, image)
    parents = [-1]*graph['nbPoints']
    parents.extend(parent)
    levels = computeLevelsAsDepth(parents)

    graph = buildFinalGraph(graph["nbPoints"], parents, adj, levels)
    drawGraphVizAttr("DCC-Hierarchy-Fig5.pdf", graph, ["level"],False)
    print("=> Result saved in file " + " 'DCC-Hierarchy-Fig5.pdf'")
    print("Done\n\n")


# Test of the naive algorithm for stack 
# Input: 
#    - file fig5-Graph.list: plain text file encoding the graph of Fig. 5
#    - file fig5-Image.png: png image encoding the weights of the graph of Fig. 5
# Output:
#    - file "DCC-Hierarchy-Fig5-Stack.pdf": pdf plot of the graph Fig. 6 c
def testAlgoStackGraphArticle():
    print("Test of the general algorithm for stacks of graphs")
    adj = readGraph("fig5-Graph.list")
    image, size = readImage("fig5-Image.png")
    stack = createGraphStackFromVertexWeightedGraph(adj,image)
    
    parent, DAGsAdj, Lvls, nbOut = directedComponentHierarchyStack(stack)
    ensureTree(parent, Lvls)
    graph = buildFinalGraph(stack["nbPoints"], parent, DAGsAdj, Lvls)
    
    drawGraphVizAttr("DCC-Hierarchy-Fig5-Stack.pdf", graph, ["level"],False)
    print("=> Result saved in file " + " 'DCC-Hierarchy-Fig5-Stack.pdf'")
    print("Done\n\n")


# Test of the semi-naive algorithm for stack 
# Input: 
#    - file fig5-Graph.list : plain text file encoding the graph of Fig. 5
#    - file fig5-Image.png: png image encoding the weights of the graph of Fig. 5
# Output:
#    - file "DCC-Hierarchy-Fig5-Stack-Fast.pdf": pdf plot of the graph Fig. 6 c
def testAlgoStackGraphArticleFast():
    print("Test of the general algorithm for stacks of graphs with optimization")
    adj = readGraph("fig5-Graph.list")
    image, size = readImage("fig5-Image.png")
    stack = createGraphStackFromVertexWeightedGraph(adj,image)

    parent, DAGsAdj, Lvls = directedComponentHierarchyStackFast(stack)
    ensureTree(parent, Lvls)
    graph = buildFinalGraph(stack["nbPoints"], parent, DAGsAdj, Lvls)
    
    drawGraphVizAttr("DCC-Hierarchy-Fig5-Stack-Fast.pdf", graph, ["level"],False)
    print("=> Result saved in file " + " 'DCC-Hierarchy-Fig5-Stack-Fast.pdf'")
    print("Done\n\n")
    

############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 

### Neurite filtering application

############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
##############################################################################

# Illustration on neurite image filtering 
# Input: 
#    - file "neurite.png": the original neurite image (Fig. 1 a)
#    - file "neurite-vesselness.png": the classification into background, blobs and vessels (Fig. 9 a)
# Output:
#    - file "Neurite-Filtered-DCC.png": the filtered image (Fig. 9 b)
#    - file "Neurite-Filtered-DCC-Regularized.png": the regularized filtered image (Fig. 1 b)
def testNeuriteFiltering():
    print("\nIllustration on neurite image filtering")
    print("Reading images")
    classeFile = "neurite-vesselness.png"
    greyFile = "neurite.png"
    
    print("Creating adjacency")
    LabelV=70   # gray level of vessels in the classification
    LabelB=140  # gray level of blobs in the classification
    
    image, size = readImage(greyFile)
    classif, size = readImage(classeFile)
    adj = createSemanticGraphNeurite(classif,size, LabelV,LabelB)

    print("Creating hierarchy")
    stack = createGraphStackFromVertexWeightedGraph(adj,image)
    # here we don't use the optimized version of the algorithm 
    # because it does no allow us to easily compute the attribute nbOut
    parent, DAGsAdj, Lvls, nbOut = directedComponentHierarchyStack(stack)
    ensureTree(parent, Lvls)
    graph = buildFinalGraph(stack["nbPoints"], parent, DAGsAdj, Lvls, nbOut)

    print("Filtering")
    def filterRule(g, n):       
        return  len(n["sucs"])<2 or  n["nbout"]>20 
        
    filterDirect(graph, filterRule)
    saveAsImage(graph,"level","Neurite-Filtered-DCC.png",size[0])    
    print("=> Result saved in file " + " 'Neurite-Filtered-DCC.png'")
    
    regularizeSelectMax(graph)
    regularizeSelectMaxHierarchical(graph)
    saveAsImage(graph,"level","Neurite-Filtered-DCC-Regularized.png",size[0])  
    print("=> Result saved in file " + " 'Neurite-Filtered-DCC-Regularized.png'")
    
    print("Done\n\n")  
  

############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 

### Heart segmentation application

############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
##############################################################################

# Illustration on heart segmentation 
# Input: 
#    - file "heart.pgm": the original heart image (Fig. 10 a)
#    - file "heart-ObjectMarker.pgm": the object marker (Fig. 10 b)
#    - file "heart-BackgroundMarker.pgm": the background marker (Fig. 10 c)
# Output:
#    - file "Heart-Segmentation-DCC.png": the segmentation result (Fig. 10 f)
def testHeartSegmentation():
    print("Illustration on heart image segmentation")
    print("Reading images")
    image, size = readImage("heart.pgm")
    markerObj, size = readImage("heart-ObjectMarker.pgm")
    markerFond, size = readImage("heart-BackgroundMarker.pgm")
    
    print("Creating adjacency")
    def weightFunction(v1,v2):
        diff=abs(v1-v2)
        if v2 <= 37 or v2 > 116: # hardcoded classification !  (Fig. 10 e)
            return (1.5)*diff
        else:
            return diff
    graph = createHeartGradientGraph(image, size, weightFunction, 8)
    
    print("Creating hierarchy")
    stack = createGraphStackFromEdgeWeightedGraph(graph)
    parent, completeGraphEdges, Lvls = directedComponentHierarchyStackFast(stack)
    graph = buildFinalGraph(stack["nbPoints"], parent, completeGraphEdges, Lvls)
    
    print("Computing attributes")
    computeAttributeMarker(graph, markerObj, 255, "marker1")  
    computeAttributeMarker(graph, markerFond, 255, "marker2")
    computeAttributeMarkerDirected(graph,"marker2")
    
    print("Filtering")
    def filterRule(g, n):
        return not(n["marker1"] and not n["marker2_directed"])
    
    filterDirect(graph, filterRule)
    regularizeSelectMax(graph)
    regularizeSelectMaxHierarchical(graph)

    print("Reconstruction")
    saveAsImage(graph,"deleted","Heart-Segmentation-DCC.png",size[0],255)    
    print("=> Result saved in file " + " 'Heart-Segmentation-DCC.png'")
    
    print("Done\n\n")      
    


############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 

### Eye fundus filtering application

############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
##############################################################################

# Create a k nearest neighbour graph
#  - image: input image
#  - size: dimensions [width, height] of image
#  - similarityFunc: a function that weights the similarity between two pixel values (the lower the more similar)
#  - k: number of neighbors to select
#  - maxD: nearest neighbors are selected in a window of size (2*maxD+1)x(2*maxD+1)
def createKNNNeighbourhoodGraph(image,size,similarityFunc,k,maxD):
    graph =  createGraph(size[0]*size[1])
    width=size[0]
    height=size[1]
    coordLinTo2D = lambda x: (x % width, x // width)
    coord2DToLin= lambda x,y: y*width+x
    dim = width * height
    for i in range(dim):
        px,py = coordLinTo2D(i)
        measures=[]
        for y in range(py-maxD,py+maxD+1,1):
            if y>=0 and y<height:
                for x in range(px-maxD,px+maxD+1,1):
                    if x>=0 and x<width:
                        if(x!=px or y!=py):
                            pos=coord2DToLin(x,y)
                            op=similarityFunc(image,size,i,pos)
                            if(op!=None):
                                measures.append((pos,op))
        measures=sorted(measures, key = lambda x: x[1])
        for j in range(min(len(measures),k)):
            createEdge(graph,i,measures[j][0])
    return graph
            


# Pseudo symilarity measure for finding the brighest pixels
#    - im: image
#    - size: dimensions of im
#    - p1: first pixel
#    - p2: second pixel
def simMax(im,size,p1,p2):
        return -im[p2]
    

# Creates the d-component hierarchy induced by the image and the adjacency. 
# Size is the dimensions [width,height] of the image.
def prepareHierarchy(adjacency,image,size):
    print("Creating hierarchy")
    stack = createGraphStackFromVertexWeightedGraph(adjacency,image)
    parent, completeGraphEdges, Lvls = directedComponentHierarchyStackFast(stack)
    graph = buildFinalGraph(stack["nbPoints"], parent, completeGraphEdges, Lvls)
    
    print("Computing attributes")
    computeAttributeLeavesCoordinates(graph, size[0])
    computeAttributeArea(graph)
    computeAttributeSimpleMoments(graph)
    computeAttributeElongationOrientation(graph)
    computeAttributeInertia(graph)
    return graph

# Retinal image filtering of one image of the DRIVE database
#    imageNumber: number of the image to process
#    method: method number between 1 and 5 (see description of the method testRetineAll)
# Input: 
#    - file "DRIVE/<imageNumber>_test_filtered.png": the preprocessed image of the DRIVE database number <imageNumber>
# Output:
#    - filtering result saved in file "DRIVE/<imageNumber>_test_filtered_result_method_<method>.png"
def filterRetine(imageNumber,method=1):

    if method==1:
        print("Filtering with regularization on D-components on a non-local asymmetric graph.")
    elif method==2:
        print("Filtering on D-components on a non-local asymmetric graph.")
    elif method==3:
        print("Filtering on S-components on a non-local asymmetric graph.")
    elif method==4:
        print("Filtering on connected components on a non-local symmetric graph (symmetrization strategy: max).")
    elif method==5:
        print("Filtering on connected components on a non-local symmetric graph (symmetrization strategy: min).")
    else :
        print("Incorrect method !")
        return
    
    numS="%02d" % (imageNumber)
    fileName=numS+"_test_filtered"
    filePath="DRIVE\\" +fileName + ".png"
    
    print("Reading image: " + filePath)
    image, size = readImage(filePath)

    print("Creating adjacency")
    
    adj = createKNNNeighbourhoodGraph(image,size,simMax,3,4)
    adj = transposeGraph(adj)
   
    if method==4:
        adj=symmetrizeGraphMax(adj)
    elif method==5:
        adj=symmetrizeGraphMin(adj)
    
    createSimpleNeighbourhoodGraph(size,4, adj)
    
    graph=prepareHierarchy(adj,image,size)


    print("Node selection")
    def filterRule_NL_Asym_DCC(g, n):       
        return    n['area']<0 or n['area_semi']<15 or  (n["inertia_semi"]<1 and n['area_semi']>=1000)  or (  n["elongation_semi"]>0.2 and n['area_semi']<1000)
    
    def filterRule_NL_Asym_SCC(g, n):       
        return    n['area']<30 or  (n["inertia"]<1.3 and n['area']>=1000)  or (  n["elongation"]>0.15 and n['area']<1000)
    
    def filterRule_NL_SymMax(g, n):       
        return    n['area']<0 or n['area_semi']<20 or  (n["inertia_semi"]<1.3 and n['area_semi']>=1000)  or (  n["elongation_semi"]>0.2 and n['area_semi']<1000)
   
    def filterRule_NL_SymMin(g, n):       
        return    n['area']<0 or n['area_semi']<15 or  (n["inertia_semi"]<1.35 and n['area_semi']>=1000)  or (  n["elongation_semi"]>0.15 and n['area_semi']<1000)
   
    if method==1:
        filterDirect(graph, filterRule_NL_Asym_DCC)
        regularizeDiscardMin(graph)
    elif method==2:
        filterDirect(graph, filterRule_NL_Asym_DCC)
    elif method==3:
        filterDirect(graph, filterRule_NL_Asym_SCC)
    elif method==4:
        filterDirect(graph, filterRule_NL_SymMax)
    elif method==5:
        filterDirect(graph, filterRule_NL_SymMin)
        
    print("Reconstruction")
    im=getAsImage(graph,"level",size[0])  
  
    saveName=fileName+"_result_method_" + str(method)+".png"  
    im.save("DRIVE\\" + saveName, "PNG")
    print("=> Result saved in file " + "DRIVE\\" +saveName)
  

# Retinal image filtering of all the images of the DRIVE database
#    method: method number between 1 and 5 (see description in the paper)
#       - method==1: Filtering with regularization on D-components on a non-local asymmetric graph
#       - method==2: Filtering on D-components on a non-local asymmetric graph
#       - method==3: Filtering on S-components on a non-local asymmetric graph
#       - method==4: Filtering on connected components on a non-local symmetric graph (symmetrization strategy: max)
#       - method==5: Filtering on connected components on a non-local symmetric graph (symmetrization strategy: min)
# Input: 
#    - files "DRIVE/XX_test_filtered.png" for XX from 00 to 20: the preprocessed image of the DRIVE database number <imageNumber>
# Output:
#    - filtering result saved in file "DRIVE/XX_test_filtered_result_method_<method>.png"  for XX from 00 to 20
def testRetineAll(method):
    for i in range(1,21):
        filterRetine(i, method)  

############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 

### Main 

############################################################################## 
############################################################################## 
############################################################################## 
############################################################################## 
##############################################################################   
    

testAlgoArticle()
 
testAlgoStackGraphArticle()

testAlgoStackGraphArticleFast()

testNeuriteFiltering()

testHeartSegmentation()

# argument => method number, see function documentation
testRetineAll(1)
